@@ -28,7 +28,10 @@ try:
     from pyghmi.ipmi.private import session
 except ImportError:
     session = None
-from pyghmi.ipmi.private import localsession
+try:
+    from pyghmi.ipmi.private import localsession
+except ImportError:
+    localsession = None
 import pyghmi.ipmi.private.util as pygutil
 import pyghmi.ipmi.sdr as sdr
 import socket
@@ -247,6 +250,18 @@ class Command(object):
                         'persistent': persistent,
                         'uefimode': uefimode}
 
+    def reseat_bay(self, bay):
+        """Request the reseat of a bay
+
+        Request the enclosure manager to reseat the system in a particular
+        bay.
+
+        :param bay: The bay identifier to reseat
+        :return:
+        """
+        self.oem_init()
+        self._oem.reseat_bay(bay)
+
     def set_power(self, powerstate, wait=False):
         """Request power state change (helper)
 
@@ -540,7 +555,7 @@ class Command(object):
         yield "System"
         if self._sdr is None:
             self._sdr = sdr.SDR(self)
-        for fruid in self._sdr.fru:
+        for fruid in sorted(self._sdr.fru):
             yield self._sdr.fru[fruid].fru_name
         self.oem_init()
         for compname in self._oem.get_oem_inventory_descriptions():
@@ -601,7 +616,7 @@ class Command(object):
         yield ("System", self._get_zero_fru())
         if self._sdr is None:
             self._sdr = sdr.SDR(self)
-        for fruid in self._sdr.fru:
+        for fruid in sorted(self._sdr.fru):
             fruinf = fru.FRU(
                 ipmicmd=self, fruid=fruid, sdr=self._sdr.fru[fruid]).info
             if fruinf is not None:
@@ -685,7 +700,7 @@ class Command(object):
             if prefixlen:
                 return _mask_to_cidr(fetchdata[1:])
             else:
-                ip = socket.inet_ntop(socket.AF_INET, fetchdata[1:])
+                ip = socket.inet_ntoa(fetchdata[1:])
                 if ip == '0.0.0.0':
                     return None
                 return ip
@@ -700,6 +715,14 @@ class Command(object):
         else:
             raise Exception("Unrecognized data format " + repr(fetchdata))
 
+    def get_system_configuration(self):
+        self.oem_init()
+        return self._oem.get_system_configuration()
+
+    def set_system_configuration(self, changeset):
+        self.oem_init()
+        return self._oem.set_system_configuration(changeset)
+
     def set_net_configuration(self, ipv4_address=None, ipv4_configuration=None,
                               ipv4_gateway=None, channel=None):
         """Set network configuration data.
@@ -740,6 +763,59 @@ class Command(object):
             cmddata = bytearray((channel, 12)) + socket.inet_aton(ipv4_gateway)
             self.xraw_command(netfn=0xc, command=1, data=cmddata)
 
+    def get_storage_configuration(self):
+        """"Get storage configuration data
+
+        Retrieves the storage configuration from the target.  Data is given
+        about disks, pools, and volumes.  When referencing something, use the
+        relevant 'cfgpath' attribute to describe it.  It is not guaranteed that
+        cfgpath will be consistent version to version, so a lookup is suggested
+        in end user applications.
+
+        :return: A pyghmi.storage.ConfigSpec object describing current config
+        """
+        self.oem_init()
+        return self._oem.get_storage_configuration()
+
+    def clear_storage_arrays(self):
+        """Remove all array and dependent volumes from the system
+
+        :return:
+        """
+        self.oem_init()
+        self._oem.clear_storage_arrays()
+
+    def remove_storage_configuration(self, cfgspec):
+        """Remove specified storage configuration from controller.
+
+        :param cfgspec: A pyghmi.storage.ConfigSpec describing what to remove
+        :return:
+        """
+        self.oem_init()
+        return self._oem.remove_storage_configuration(cfgspec)
+
+    def apply_storage_configuration(self, cfgspec=None):
+        """Evaluate a configuration for validity
+
+        This will check if configuration is currently available and, if given,
+        whether the specified cfgspec can be applied.
+        :param cfgspec: A pyghmi.storage.ConfigSpec describing desired oonfig
+        :return:
+        """
+        self.oem_init()
+        return self._oem.apply_storage_configuration(cfgspec)
+
+    def check_storage_configuration(self, cfgspec=None):
+        """Evaluate a configuration for validity
+
+        This will check if configuration is currently available and, if given,
+        whether the specified cfgspec can be applied.
+        :param cfgspec: A pyghmi.storage.ConfigSpec describing desired oonfig
+        :return:
+        """
+        self.oem_init()
+        return self._oem.check_storage_configuration(cfgspec)
+
     def get_net_configuration(self, channel=None, gateway_macs=True):
         """Get network configuration data
 
@@ -901,8 +977,7 @@ class Command(object):
         rsp = self.xraw_command(netfn=0xc, command=2, data=rqdata)
         if ord(rsp['data'][2]) & 0b11110000 == 0:
             destinfo['address_format'] = 'ipv4'
-            destinfo['address'] = socket.inet_ntop(socket.AF_INET,
-                                                   rsp['data'][4:8])
+            destinfo['address'] = socket.inet_ntoa(rsp['data'][4:8])
         elif ord(rsp['data'][2]) & 0b11110000 == 0b10000:
             destinfo['address_format'] = 'ipv6'
             destinfo['address'] = socket.inet_ntop(socket.AF_INET6,
@@ -1024,7 +1099,7 @@ class Command(object):
         if ip is not None:
             destdata = bytearray((channel, 19, destination))
             try:
-                parsedip = socket.inet_pton(socket.AF_INET, ip)
+                parsedip = socket.inet_aton(ip)
                 destdata.extend((0, 0))
                 destdata.extend(parsedip)
                 destdata.extend(b'\x00\x00\x00\x00\x00\x00')
@@ -1715,14 +1790,14 @@ class Command(object):
         self.set_user_password(uid, mode)
         return True
 
-    def get_firmware(self):
+    def get_firmware(self, components=()):
         """Retrieve OEM Firmware information
         """
         self.oem_init()
         mcinfo = self.xraw_command(netfn=6, command=1)
         bmcver = '{0}.{1}'.format(
             ord(mcinfo['data'][2]), hex(ord(mcinfo['data'][3]))[2:])
-        return self._oem.get_oem_firmware(bmcver)
+        return self._oem.get_oem_firmware(bmcver, components)
 
     def get_capping_enabled(self):
         """Get PSU based power capping status
@@ -1765,6 +1840,21 @@ class Command(object):
         self.oem_init()
         return self._oem.get_graphical_console()
 
+    def update_firmware(self, file, data=None, progress=None, bank=None):
+        """Send file to BMC to perform firmware update
+
+         :param filename:  The filename to upload to the target BMC
+         :param data:  The payload of the firmware.  Default is to read from
+                       specified filename.
+         :param progress:  A callback that will be given a dict describing
+                           update process.  Provide if
+         :param bank: Indicate a target 'bank' of firmware if supported
+        """
+        self.oem_init()
+        if progress is None:
+            progress = lambda x: True
+        return self._oem.update_firmware(file, data, progress, bank)
+
     def attach_remote_media(self, url, username=None, password=None):
         """Attach remote media by url
 
@@ -1778,8 +1868,30 @@ class Command(object):
         :param password: Password for endpoint to use when accessing the URL.
         """
         self.oem_init()
-        self._oem.attach_remote_media(url, username, password)
+        return self._oem.attach_remote_media(url, username, password)
 
     def detach_remote_media(self):
         self.oem_init()
-        self._oem.detach_remote_media()
+        return self._oem.detach_remote_media()
+
+    def upload_media(self, filename, progress=None):
+        """Upload a file to be hosted on the target BMC
+
+        This will upload the specified data to
+        the BMC so that it will make it available to the system as an emulated
+        USB device.
+
+        :param filename: The filename to use, the basename of the parameter
+                         will be given to the bmc.
+        :param filename: Optional callback for progress updates
+        """
+        self.oem_init()
+        return self._oem.upload_media(filename, progress)
+
+    def list_media(self):
+        """List attached remote media
+
+        :returns: An iterable list of attached media
+        """
+        self.oem_init()
+        return self._oem.list_media()
