@@ -1,6 +1,6 @@
 # vim: tabstop=4 shiftwidth=4 softtabstop=4
 
-# Copyright 2015-2016 Lenovo
+# Copyright 2015-2017 Lenovo
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -154,6 +154,8 @@ class OEMHandler(generic.OEMHandler):
             self.immhandler = imm.XCCClient(ipmicmd)
         elif self.has_imm:
             self.immhandler = imm.IMMClient(ipmicmd)
+        elif self.is_fpc:
+            self.smmhandler = nextscale.SMMClient(ipmicmd)
 
     @property
     def _megarac_eth_index(self):
@@ -163,6 +165,31 @@ class OEMHandler(generic.OEMHandler):
             self._mrethidx = rsp['data'][0]
         return self._mrethidx
 
+    def remove_storage_configuration(self, cfgspec):
+        if self.has_xcc:
+            return self.immhandler.remove_storage_configuration(cfgspec)
+        return super(OEMHandler, self).remove_storage_configuration()
+
+    def clear_storage_arrays(self):
+        if self.has_xcc:
+            return self.immhandler.clear_storage_arrays()
+        return super(OEMHandler, self).clear_storage_ararys()
+
+    def apply_storage_configuration(self, cfgspec):
+        if self.has_xcc:
+            return self.immhandler.apply_storage_configuration(cfgspec)
+        return super(OEMHandler, self).apply_storage_configuration()
+
+    def check_storage_configuration(self, cfgspec):
+        if self.has_xcc:
+            return self.immhandler.check_storage_configuration(cfgspec)
+        return super(OEMHandler, self).get_storage_configuration()
+
+    def get_storage_configuration(self):
+        if self.has_xcc:
+            return self.immhandler.get_storage_configuration()
+        return super(OEMHandler, self).get_storage_configuration()
+
     def get_video_launchdata(self):
         if self.has_tsm:
             return self.get_tsm_launchdata()
@@ -245,10 +272,17 @@ class OEMHandler(generic.OEMHandler):
                 event['component_type_id'] == 13):
             event['component'] += ' {0}'.format(evdata[1] & 0b11111)
 
+    def reseat_bay(self, bay):
+        if self.is_fpc:
+            return self.smmhandler.reseat_bay(bay)
+        return super(OEMHandler, self).reseat_bay(bay)
+
     def get_ntp_enabled(self):
         if self.has_tsm:
             ntpres = self.ipmicmd.xraw_command(netfn=0x32, command=0xa7)
             return ntpres['data'][0] == '\x01'
+        elif self.is_fpc:
+            return self.smmhandler.get_ntp_enabled(self._fpc_variant)
         return None
 
     def get_ntp_servers(self):
@@ -258,6 +292,8 @@ class OEMHandler(generic.OEMHandler):
             srvs.append(ntpres['data'][1:129].rstrip('\x00'))
             srvs.append(ntpres['data'][129:257].rstrip('\x00'))
             return srvs
+        if self.is_fpc:
+            return self.smmhandler.get_ntp_servers()
         return None
 
     def set_ntp_enabled(self, enabled):
@@ -269,6 +305,9 @@ class OEMHandler(generic.OEMHandler):
                 self.ipmicmd.xraw_command(
                     netfn=0x32, command=0xa8, data=(3, 0), timeout=15)
             return True
+        if self.is_fpc:
+            self.smmhandler.set_ntp_enabled(enabled)
+            return True
         return None
 
     def set_ntp_server(self, server, index=0):
@@ -279,17 +318,23 @@ class OEMHandler(generic.OEMHandler):
             cmddata += server.ljust(128, '\x00')
             self.ipmicmd.xraw_command(netfn=0x32, command=0xa8, data=cmddata)
             return True
+        elif self.is_fpc:
+            if not 0 <= index <= 2:
+                raise pygexc.InvalidParameterValue(
+                    'SMM supports indexes 0 through 2')
+            self.smmhandler.set_ntp_server(server, index)
+            return True
         return None
 
     @property
     def is_fpc(self):
         """True if the target is a Lenovo nextscale fan power controller
         """
-        fpc_id = (19046, 32, 1063)
+        fpc_ids = ((19046, 32, 1063), (20301, 32, 462))
         smm_id = (19046, 32, 1180)
         currid = (self.oemid['manufacturer_id'], self.oemid['device_id'],
                   self.oemid['product_id'])
-        if currid == fpc_id:
+        if currid in fpc_ids:
             self._fpc_variant = 6
         elif currid == smm_id:
             self._fpc_variant = 2
@@ -340,16 +385,25 @@ class OEMHandler(generic.OEMHandler):
             for name in nextscale.get_sensor_names(self._fpc_variant):
                 yield nextscale.get_sensor_reading(name, self.ipmicmd,
                                                    self._fpc_variant)
+        elif self.has_imm:
+            for name in self.immhandler.get_oem_sensor_names(self.ipmicmd):
+                yield self.immhandler.get_oem_sensor_reading(name,
+                                                             self.ipmicmd)
 
     def get_sensor_descriptions(self):
         if self.is_fpc:
             return nextscale.get_sensor_descriptions(self._fpc_variant)
+        elif self.has_imm:
+            return self.immhandler.get_oem_sensor_descriptions(self.ipmicmd)
         return ()
 
     def get_sensor_reading(self, sensorname):
         if self.is_fpc:
             return nextscale.get_sensor_reading(sensorname, self.ipmicmd,
                                                 self._fpc_variant)
+        elif self.has_imm:
+            return self.immhandler.get_oem_sensor_reading(sensorname,
+                                                          self.ipmicmd)
         return ()
 
     def get_inventory_of_component(self, component):
@@ -494,6 +548,9 @@ class OEMHandler(generic.OEMHandler):
             except (AttributeError, KeyError, IndexError):
                 pass
             return fru
+        elif self.is_fpc == 2:  # SMM variant
+            fru['oem_parser'] = 'lenovo'
+            return self.smmhandler.process_fru(fru)
         else:
             fru['oem_parser'] = None
             return fru
@@ -538,13 +595,13 @@ class OEMHandler(generic.OEMHandler):
         self._hasimm = (rdata[1] & 1 == 1) or (rdata[1] & 16 == 16)
         return self._hasimm
 
-    def get_oem_firmware(self, bmcver):
+    def get_oem_firmware(self, bmcver, components):
         if self.has_tsm:
             command = firmware.get_categories()["firmware"]
             rsp = self.ipmicmd.xraw_command(**command["command"])
             return command["parser"](rsp["data"])
         elif self.has_imm:
-            return self.immhandler.get_firmware_inventory(bmcver)
+            return self.immhandler.get_firmware_inventory(bmcver, components)
         elif self.is_fpc:
             return nextscale.get_fpc_firmware(bmcver, self.ipmicmd,
                                               self._fpc_variant)
@@ -828,6 +885,27 @@ class OEMHandler(generic.OEMHandler):
                 else:
                     raise
 
+    def update_firmware(self, filename, data=None, progress=None, bank=None):
+        if self.has_xcc:
+            return self.immhandler.update_firmware(
+                filename, data=data, progress=progress, bank=bank)
+        if self.is_fpc:
+            return self.smmhandler.update_firmware(
+                filename, data=data, progress=progress, bank=bank)
+        return super(OEMHandler, self).update_firmware(filename, data=data,
+                                                       progress=progress,
+                                                       bank=bank)
+
+    def get_system_configuration(self):
+        if self.has_imm or self.has_xcc:
+            return self.immhandler.get_system_configuration()
+        return super(OEMHandler, self).get_system_configuration()
+
+    def set_system_configuration(self, changeset):
+        if self.has_imm or self.has_xcc:
+            return self.immhandler.set_system_configuration(changeset)
+        return super(OEMHandler, self).set_system_configuration(changeset)
+
     def detach_remote_media(self):
         if self.has_imm:
             self.immhandler.detach_remote_media()
@@ -835,3 +913,13 @@ class OEMHandler(generic.OEMHandler):
             self.ipmicmd.xraw_command(
                 netfn=0x32, command=0x9f, data=(8, 10, 0, 0))
             self.ipmicmd.xraw_command(netfn=0x32, command=0x9f, data=(8, 11))
+
+    def upload_media(self, filename, progress):
+        if self.has_xcc:
+            return self.immhandler.upload_media(filename, progress)
+        return super(OEMHandler, self).upload_media(filename, progress)
+
+    def list_media(self):
+        if self.has_xcc:
+            return self.immhandler.list_media()
+        return super(OEMHandler, self).list_media()
