@@ -291,6 +291,7 @@ class Session(object):
     """
     bmc_handlers = {}
     waiting_sessions = {}
+    initting_sessions = {}
     keepalive_sessions = {}
     peeraddr_to_nodes = {}
     iterwaiters = []
@@ -440,8 +441,15 @@ class Session(object):
                     forbidsock.append(self.socket)
             if trueself:
                 return trueself
+            i = cls.initting_sessions.get(
+                (bmc, userid, password, port, kg), False)
+            if i:
+                i.initialized = True
+                i.logging = True
+                return i
             self = object.__new__(cls)
             self.forbidsock = forbidsock
+            cls.initting_sessions[(bmc, userid, password, port, kg)] = self
             return self
 
     def __init__(self,
@@ -475,7 +483,6 @@ class Session(object):
         # queue of events denoting line to run a cmd
         self.evq = collections.deque([])
         self.bmc = bmc
-        self.broken = False
         # a private queue for packets for which this session handler
         # is destined to receive
         self.pktqueue = collections.deque([])
@@ -547,7 +554,7 @@ class Session(object):
                 myport = self.socket.getsockname()[1]
                 for sockaddr in self.allsockaddrs:
                     if (sockaddr in Session.bmc_handlers and
-                            myport in Session.bmc_hansdlers[sockaddr]):
+                            myport in Session.bmc_handlers[sockaddr]):
                         del Session.bmc_handlers[sockaddr][myport]
                         if Session.bmc_handlers[sockaddr] == {}:
                             del Session.bmc_handlers[sockaddr]
@@ -707,20 +714,14 @@ class Session(object):
             incrementtime += 1
         return cumulativetime + 1
 
-    def _cmdwait(self, waitall=False):
+    def _cmdwait(self):
         while self._isincommand():
-            if waitall:
-                Session.wait_for_rsp(self._isincommand())
-            else:
-                _io_wait(self._isincommand(), self.sockaddr, self.evq)
+            _io_wait(self._isincommand(), self.sockaddr, self.evq)
 
-    def awaitresponse(self, retry, waitall=False):
+    def awaitresponse(self, retry):
         while retry and self.lastresponse is None and self.logged:
             timeout = self.expiration - _monotonic_time()
-            if waitall:
-                Session.wait_for_rsp(timeout)
-            else:
-                _io_wait(timeout, self.sockaddr)
+            _io_wait(timeout, self.sockaddr)
             while self.iterwaiters:
                 waiter = self.iterwaiters.pop()
                 waiter({'success': True})
@@ -739,15 +740,18 @@ class Session(object):
                     retry=True,
                     delay_xmit=None,
                     timeout=None,
-                    waitall=False):
+                    callback=None):
         if not self.logged:
             raise exc.IpmiException('Session no longer connected')
-        self._cmdwait(waitall)
+        self._cmdwait()
         if not self.logged:
             raise exc.IpmiException('Session no longer connected')
         self.incommand = _monotonic_time() + self._getmaxtimeout()
         self.lastresponse = None
-        self.ipmicallback = self._generic_callback
+        if callback is None:
+            self.ipmicallback = self._generic_callback
+        else:
+            self.ipmicallback = callback
         self._send_ipmi_net_payload(netfn, command, data,
                                     bridge_request=bridge_request,
                                     retry=retry, delay_xmit=delay_xmit,
@@ -757,13 +761,16 @@ class Session(object):
             timeout = None
         else:  # if not retry, give it a second before surrending
             timeout = 1
+        if callback:
+            # caller *must* clean up self.incommand and self.evq
+            return
         # The event loop is shared amongst pyghmi session instances
         # within a process.  In this way, synchronous usage of the interface
         # plays well with asynchronous use.  In fact, this produces the
         # behavior of only the constructor needing a callback.  From then on,
         # synchronous usage of the class acts in a greenthread style governed
         # by order of data on the network
-        self.awaitresponse(retry, waitall)
+        self.awaitresponse(retry)
         lastresponse = self.lastresponse
         self.incommand = False
         while self.evq:
@@ -1076,7 +1083,7 @@ class Session(object):
 
         This watches for any activity on IPMI handles and handles registered
         by register_handle_callback.  Callers are satisfied in the order that
-        packets return from nework, not in the order of calling.
+        packets return from network, not in the order of calling.
 
         :param timeout: Maximum time to wait for data to come across.  If
                         unspecified, will autodetect based on earliest timeout
@@ -1194,16 +1201,26 @@ class Session(object):
         except KeyError:
             pass
 
+    def _keepalive_wrapper(self, callback):
+        # generates a wrapped keepalive to cleanup session state
+        # and call callback if appropriate
+        def _keptalive(response):
+            self._generic_callback(response)
+            response = self.lastresponse
+            self.incommand = False
+            while self.evq:
+                self.evq.popleft().set()
+            if callback:
+                callback(response)
+
+        return _keptalive
+
     def _keepalive(self):
         """Performs a keepalive to avoid idle disconnect
         """
         try:
-            if not self._customkeepalives:
-                if self.incommand:
-                    # if currently in command, no cause to keepalive
-                    return
-                self.raw_command(netfn=6, command=1, waitall=True)
-            else:
+            keptalive = False
+            if self._customkeepalives:
                 kaids = list(self._customkeepalives.keys())
                 for keepalive in kaids:
                     try:
@@ -1215,8 +1232,18 @@ class Session(object):
                         # raw command ultimately caused a keepalive to
                         # deregister
                         continue
-                    cmd['waitall'] = True
-                    callback(self.raw_command(**cmd))
+                    if callable(cmd):
+                        cmd()
+                        continue
+                    keptalive = True
+                    cmd['callback'] = self._keepalive_wrapper(callback)
+                    self.raw_command(**cmd)
+            if not keptalive:
+                if self.incommand:
+                    # if currently in command, no cause to keepalive
+                    return
+                self.raw_command(netfn=6, command=1,
+                                 callback=self._keepalive_wrapper(None))
         except exc.IpmiException:
             self._mark_broken()
 
@@ -1690,6 +1717,12 @@ class Session(object):
                         Session.bmc_handlers[sockaddr] = {}
                     Session.bmc_handlers[sockaddr][myport] = self
                     _io_sendto(self.socket, self.netpacket, sockaddr)
+                try:
+                    del Session.initting_sessions[(self.bmc, self.userid,
+                                                   self.password, self.port,
+                                                   self.kgo)]
+                except KeyError:
+                    pass
             except socket.gaierror:
                 raise exc.IpmiException(
                     "Unable to transmit to specified address")
